import sets
import copy


#LOS pre-calculations
VIEW_DROPOFF = .2
FENCE_OPACITY = 0.5

def PRE_GENERATE_SIGHT_RAYS(View_Dropoff):
    Ray_Length = int(1/View_Dropoff)
    LOS_quadrant = []

    #We calculate the ray endpoints for the circle AT
    # the maximum view distance, and the diags from
    # the circle one shorter to prevent raster holes
    
    for i in range(Ray_Length):
        LOS_quadrant.append((i,Ray_Length-i))
        
    for i in range(1,Ray_Length):
        LOS_quadrant.append((i,(Ray_Length -1)-i))

        
    LOS_quad_rays = []
    for dx,dy in LOS_quadrant:
        """
        Brensenham line algorithm (modified)
        Used to calculate the rays to each destination cell in the quadrant
        """
        x,y = (0,0)
        steep = 0
        coords = []
        if dy > dx:
            steep = 1
            dx,dy = dy,dx
        d = (2 * dy) - dx
        for i in range(0,dx):
            while d >= 0:
                y = y + 1
                d = d - (2 * dx)
            x = x + 1
            d = d + (2 * dy)
            if steep: coords.append((y,x))
            else: coords.append((x,y))
        LOS_quad_rays.append(coords)

    LOS_ray_circle = []
    for ray in LOS_quad_rays:
        Quad1ray = []
        Quad2ray = []
        Quad3ray = []
        Quad4ray = []
        for x,y in ray:
            Quad1ray.append((x,y))
            Quad2ray.append((-y,x))
            Quad3ray.append((-x,-y))
            Quad4ray.append((y,-x))
        LOS_ray_circle.extend([Quad1ray, Quad2ray, Quad3ray, Quad4ray])
    return LOS_ray_circle

LOS_RAYS = PRE_GENERATE_SIGHT_RAYS(VIEW_DROPOFF)


class GameState:
    """
    Object representing the current state of the game at any time.
    """
    def __init__(self):
        self.playfield = None
        self.counter = 0
        self.person = None

    def create_playfield(self):
        self.playfield = playfield()
        self.OldMan = filter(lambda ob: ob.get_type() == PF_OLD_MAN, self.playfield.objects)[0]
        
    def simulate(self):
        if self.playfield is None:
            self.create_playfield()
        self.playfield.simulate(self)
        if self.person is None:
            if self.counter < 30:
                self.counter +=1
            else:
                self.counter = 0
                # spawn a person!
                c = self.playfield.get_cell((9,0))
                t = self.playfield.get_cell((0,0))
                self.person = person(c,t)
                c.add_object(self.person)
        elif not self.person.exists:
                self.person = None
        
    def update(self, dt):
        '''
        update the simulation
        dt is time delta in seconds since last update
        '''
        self.simulate()

"""
Initial Playfield setup is as follows (just prof of concept)

Object types:

hh = House, Impassible, blocks LOS
ww = Window, Impassible, blocks LOS
        certian game objects can be placed in the window and can be "Seen"
dd = Door, Impassible, blocks LOS
ss = Sidewalk, Passable, does not block LOS, easier to cross than "Grass"
ff = Fence, Impassible, blocks LOS, does not occupy entire cell
mm = Mailbox, Impassible, does not block LOS
xx = Shed, Impassible, Blocks LOS, lots of fun stuff can be found in the shed
oo = Old Man, ???
   = (unmarked) Grass, Passable does not block LOS, harder to cross than "Sidewalk"

  0  1  2  3  4  5  6  7  8  9
  -- -- -- -- -- -- -- -- -- --
9|  |xx|hh|hh|hh|hh|hh|hh|hh|  |
 |  |xx|hh|hh|hh|hh|hh|hh|hh|  |
  -- -- -- -- -- -- -- -- -- --
8|  |  |hh|ww|hh|dd|hh|ww|hh|  |
 |  |ff|hh|ww|hh|dd|hh|ww|hh|  |
  -- -- -- -- -- -- -- -- -- --
7|  |  |  |  |  |ss|  |  |  |  |
 |  |  |  |  |  |ss|  |  |  |  |
  -- -- -- -- -- -- -- -- -- --
6|  |  |  |oo|  |ss|  |  |  |  |
 |  |  |  |oo|  |ss|  |  |  |  |
  -- -- -- -- -- -- -- -- -- --
5|  |  |  |  |  |ss|  |  |  |  |
 |  |  |  |  |  |ss|  |  |  |  |
  -- -- -- -- -- -- -- -- -- --
4|  |  |  |  |  |ss|  |  |  |  |
 |  |  |  |  |  |ss|  |  |  |  |
  -- -- -- -- -- -- -- -- -- --
3|  |  |  |  |  |ss|  |  |  |  |
 |  |  |  |  |  |ss|  |  |  |  |
  -- -- -- -- -- -- -- -- -- --
2|  |  |  |  |  |ss|  |  |  |  |
 |  |  |  |  |  |ss|  |  |  |  |
  -- -- -- -- -- -- -- -- -- --
1|  |  |  |mm|  |ss|  |  |  |  |
 |ff|ff|ff|mm|ff|ss|ff|ff|ff|ff|
  -- -- -- -- -- -- -- -- -- --
0|ss|ss|ss|ss|ss|ss|ss|ss|ss|ss|
 |ss|ss|ss|ss|ss|ss|ss|ss|ss|ss|
  -- -- -- -- -- -- -- -- -- --
"""

PLAYFIELD_X_SIZE = 10;
PLAYFIELD_Y_SIZE = 10;

# Basic enum of initial types
# We will probably want something more robust longterm
PF_TYPES = PF_EMPTY, PF_HOUSE, PF_WINDOW, PF_DOOR, PF_SIDEWALK, PF_MAILBOX, PF_SHED, PF_OLD_MAN, PF_FLOWERBED, PF_PERSON, PF_BUSH = range(1,12)

D_DIRECTIONS = D_UP, D_LEFT, D_DOWN, D_RIGHT = range(4)

AI_GOAL_TYPES = AI_G_APPROACH, AI_G_REACH, AI_G_AVOID = range(3)
AI_STATE = AI_S_NORMAL, AI_S_EXCAPE = range(2)

# A dictionary of Cordinate/(type,parameter) pairs defining the inital playfield setup
PLAYFIELD_INITIAL_SETUP = {
    (3,1):PF_MAILBOX,
    (0,0):PF_SIDEWALK,
    (1,0):PF_SIDEWALK,
    (2,0):PF_SIDEWALK,
    (3,0):PF_SIDEWALK,
    (4,0):PF_SIDEWALK,
    (5,0):PF_SIDEWALK,
    (6,0):PF_SIDEWALK,
    (7,0):PF_SIDEWALK,
    (8,0):PF_SIDEWALK,
    (9,0):PF_SIDEWALK,
    (5,1):PF_SIDEWALK,
    (5,2):PF_SIDEWALK,
    (5,3):PF_SIDEWALK,
    (5,4):PF_SIDEWALK,
    (5,5):PF_SIDEWALK,
    (5,6):PF_SIDEWALK,
    (5,7):PF_SIDEWALK,
    (3,1):PF_MAILBOX,
    (5,8):PF_DOOR,
    (3,8):PF_WINDOW,
    (7,8):PF_WINDOW,
    (2,8):PF_HOUSE,
    (4,8):PF_HOUSE,
    (6,8):PF_HOUSE,
    (8,8):PF_HOUSE,
    (2,9):PF_HOUSE,
    (3,9):PF_HOUSE,
    (4,9):PF_HOUSE,
    (5,9):PF_HOUSE,
    (6,9):PF_HOUSE,
    (7,9):PF_HOUSE,
    (8,9):PF_HOUSE,
    (1,9):PF_SHED,
    (3,6):PF_OLD_MAN,
}

# Pairs of coordinates that a fence blocks between
PLAYFIELD_INITIAL_FENCES = [
    ( (0, 1), (0, 0) ),
    ( (1, 1), (1, 0) ),
    ( (2, 1), (2, 0) ),
    ( (4, 1), (4, 0) ),
    ( (6, 1), (6, 0) ),
    ( (7, 1), (7, 0) ),
    ( (8, 1), (8, 0) ),
    ( (9, 1), (9, 0) ),
    ( (9, 1), (9, 0) ),
    ( (1, 8), (1, 7) ),
]

def obj_allowed_in_cell(obj, cell):
    obj_type = type(obj)
    return obj_type_allowed_in_cell(obj_type, cell)

def obj_type_allowed_in_cell(obj_type, cell):
    #check to make sure that placing the object in the cell would not violate
    # any gamerules
    if cell is None:
        return False
    
    ANY = "ANY"

    # a list of funamental object types that are incompatible
    solids = [PF_HOUSE, PF_WINDOW, PF_DOOR, PF_MAILBOX, PF_BUSH]
    terrains = [PF_SIDEWALK, PF_FLOWERBED]
    
    solids_and_terrains = []
    solids_and_terrains.extend(solids)
    solids_and_terrains.extend(terrains)
    
    exclusive_dict = {
        old_man_obj:solids
    }
    
    for solid_class in [house_obj, window_obj, door_obj, mailbox_obj]:
        exclusive_dict[solid_class] = ANY
        
    for plantable_solid_class in [bush_obj]:
        exclusive_dict[plantable_solid_class] = solids_and_terrains
    
    for terrain_class in [sidewalk_obj, flowerbed_obj]:
        exclusive_dict[terrain_class] = solids_and_terrains
    
    if (obj_type in exclusive_dict.keys()):
        forbidden = exclusive_dict[obj_type]
        if (forbidden == ANY) and (len(cell.get_objects()) > 0):
            return False
        for target_obj in cell.get_objects():
            if target_obj.get_type() in forbidden:
                return False

    #add more rules as things get complicated
    return True

class game_obj(object):
    """
    Crude generic object class that is used to represent everything in the game, for now.
    """
    def __init__(self, game_obj_type, current_cell):
        self.game_obj_type = game_obj_type
        #make sure it's legal for this object ot exist here
        assert(obj_type_allowed_in_cell(type(self), current_cell))
        self.current_cell = current_cell

        self.opacity = 1
        self.visibility = 1
        self.passable = False
        self.exists = True

    def clear_existance(self):
        self.exists = False

    def set_existance(self):
        self.exists = True

    def get_type(self):
        return self.game_obj_type

    def transplant(self, new_cell):
        """
        Instantly transplants an object from its current cell to another.
        """
        assert( obj_allowed_in_cell(self, new_cell) )
        assert(self.current_cell.remove_object(self))
        self.current_cell = new_cell
        assert(self.current_cell.add_object(self))

    def get_location(self):
        return self.current_cell

    def move(self, new_cell):
        """
        Attempts to move the object from its current cell to another. Returns
        true if the move succeded, false otherwise.

        Note that move respects the passability of objects and fences, unlike
        transplant.
        """
        if obj_allowed_in_cell(self, new_cell):
            if self.current_cell.has_fence_between(new_cell):
                return False

            for new_cell_obj in new_cell.get_objects():
                if not new_cell_obj.passable:
                    return False

            assert(self.current_cell.remove_object(self))
            self.current_cell = new_cell
            assert(self.current_cell.add_object(self))
            return True

        return False


    def get_opacity(self):
        """
        Used in calculating Line of Sight (LOS)
        opacity is the multiplier by which the visibility of an object is
        changed by passing through the same cell as this object
        An opacity of 1 means this object does not block LOS
        An opacity of 0 means this object completly blocks LOS
        """
        return self.opacity

    def get_visibility(self):
        """
        Used in calculating Line of Sight (LOS)
        Represents the base visibility of an object
        A visibility of 1 means this object is completly visible
        a visibility of 0 means this object is completly invisible
        """
        return self.visibility

    def get_Field_Of_View(self, Threshold = .05):
        """
        Basic function for determinig LOS
        Radiates outward from the object and determines an "Opacity" to each cell.
        Opacity is based on a simple transit cost with a .8 dropoff
        for each cell, with additional dropoff based on the most opaque object
        in each cell.

        this algorithm is VERY sub-optimal, feel free to have fun optimizing it
        currently it appears to take 1/100th of a second to generate the object
        visibility list
        """

        # We start by generating a matrix of all the cells we can see, these matricies are
        # pairs of cells and a floating point number representing their visibility and are
        # generated by the host cell
        
        cells = self.current_cell.get_LOS_matrix(Threshold)
        
         
        # we should know all the cells we can see, and the visibility to them
        # now we just iterate throught them and generate a dictionary of cells/object list
        fov = {}
        for cell in cells.keys():
            cell_objs = cell.get_objects()
            visible_objs = []
            for obj in cell_objs:
                visibility = cells[cell] * obj.get_visibility()
                if visibility > Threshold:
                    visible_objs.append(obj)
            fov[cell] = visible_objs
        return fov

    def get_passable(self):
        """
        Returns if the object is passable or not
        """
        return self.passable

    def simulate(self, gamestate):
        pass

class house_obj(game_obj):
    """
    A peice of the house, dosen't do much really
    """
    def __init__(self, current_cell):
        game_obj.__init__(self, PF_HOUSE, current_cell)
        self.opacity = 0

class window_obj(game_obj):
    """
    A window, you can see stuff in it
    """
    def __init__(self, current_cell):
        game_obj.__init__(self, PF_WINDOW, current_cell)
        self.opacity = 0
        self.in_window = []

    def display(self, thing):
        self.in_window.append(thing)
        return True

    def remove(self, thing):
        if thing in self.in_window:
            self.in_window.remove(thing)
            return True
        return False

class door_obj(game_obj):
    """
    A door, dosen't even open right now
    """
    def __init__(self, current_cell):
        game_obj.__init__(self, PF_DOOR, current_cell)
        self.opacity = 0
        self.passable = False

class sidewalk_obj(game_obj):
    """
    A sidewalk, you walk on it
    """
    def __init__(self, current_cell):
        game_obj.__init__(self, PF_SIDEWALK, current_cell)
        self.passable = True

class mailbox_obj(game_obj):
    """
    A mailbox, pity you will never receve any mail in it
    """
    def __init__(self, current_cell):
        game_obj.__init__(self, PF_MAILBOX, current_cell)
        self.opacity = .5

class bush_obj(game_obj):
    """
    A bush
    """
    def __init__(self, current_cell):
        game_obj.__init__(self, PF_BUSH, current_cell)
        self.opacity = .4
                
class shed_obj(game_obj):
    """
    A shed, there is lots of fun stuff in the shed
    """
    def __init__(self, current_cell):
        game_obj.__init__(self, PF_SHED, current_cell)
        self.opacity = 0
        #The sheds inventory
        self.objects = []

    def add_object(self, target_obj):
        self.objects.append(target_obj)

    def remove_object(self, target_obj):
        if target_obj in self.objects:
            self.objects.remove(target_obj)
            return True
        return False

class old_man_obj(game_obj):
    """
    GET OFF MY LAWN!
    """
    def __init__(self, current_cell):
        game_obj.__init__(self, PF_OLD_MAN, current_cell)
        self.opacity = .2
        #The old mans inventory
        self.objects = []

    def add_object(self, target_obj):
        self.objects.append(target_obj)

    def remove_object(self, target_obj):
        if target_obj in self.objects:
            self.objects.remove(target_obj)
            return True
        return False

    def simulate(self, gamestate):
        pass
    
    def move_up(self):
        self.move(self.current_cell.get_adjacent_cell_in_direction(D_UP))
        
    def move_down(self):
        self.move(self.current_cell.get_adjacent_cell_in_direction(D_DOWN))
        
    def move_left(self):
        self.move(self.current_cell.get_adjacent_cell_in_direction(D_LEFT))
        
    def move_right(self):
        self.move(self.current_cell.get_adjacent_cell_in_direction(D_RIGHT))


class flowerbed_obj(game_obj):
    """
    A beatiful flowerbed.
    
    Takes a few simulation ticks to bloom after planting.
    Trampled progressly by AI objects.
    """
    def __init__(self, current_cell):
        game_obj.__init__(self, PF_FLOWERBED, current_cell)
        self.passable = True
        self.bloom = 0
        self.opacity = 0.8
        
    def simulate(self, gamestate):
        super(flowerbed_obj, self).simulate(gamestate)
        if self.bloom < 0:
            self.current_cell.remove_object(self)
        elif self.bloom < 1:
            self.bloom += 0.05
        
        if filter(lambda obj: obj.get_type() == PF_PERSON, self.current_cell.get_objects()):
            self.bloom -= 0.25



class AI_obj(game_obj):
    class Mental_Cell():
        def __init__(self, cell, objects):
            self.cell = cell
            self.objects = objects

    class Goal():
        def __init__(self, Goal_type, Location, Strength, Sustain = 3):
            self.Goal_type = Goal_type
            self.Location = Location
            self.Strength = Strength
            self.age = 0
            self.Sustain = 3

        def tick(self):
            self.age += 1

        def reset_age(self):
            self.age = 0

        def fufill(self):
            if self.Sustain <= 0:
                return True
            else:
                self.Sustain -= 1
                return False

        def get_strength(self):
            return self.Strength - (.2 * self.age)

        def check_goals(self, check_cell):
            goal_cell = self.Location.cell
            if self.Goal_type is AI_G_REACH:
                if goal_cell is check_cell:
                    return True
                else:
                    return False
            elif self.Goal_type is AI_G_APPROACH:
                ajunct = check_cell.get_adjacent_cell_map().values()
                if goal_cell in ajunct:
                    return True
                else:
                    return False
            return False

        def calc_heluristic(self, check_cell):
            goal_heluristic = 0
            if self.Goal_type is AI_G_AVOID:
                goal_heluristic = -check_cell.get_distance(self.Location.cell)
            else:
                goal_heluristic = check_cell.get_distance(self.Location.cell)
            return goal_heluristic * self.get_strength()
    
    def __init__(self, pf_type, current_cell, terminal_cell):
        game_obj.__init__(self, pf_type, current_cell)
        self.known_world = {}
        self.known_world[terminal_cell] = self.Mental_Cell(terminal_cell, [])
        self.terminal_goal = self.Goal(AI_G_REACH, self.known_world[terminal_cell], 1)
        self.current_goals = {} #obj: (goal_type, location, strength)
        self.last_path = None
        self.goals_changed = False
        self.exhuasted_goals = []
        self.mental_state = AI_S_NORMAL

    def check_goals(self, cell, goals):
        for goal in goals:
            if goal.check_goals(cell):
                return True
        return False

    def calc_heluristic(self, cell, goals):
        heluristic = 0
        for goal in goals:
            heluristic += goal.calc_heluristic(cell)
        return heluristic

    def pathfind(self, goals):
        if len(goals) <= 0:
            #if we dont' have any goals, we can't pathfind!
            return []
        
        class possible_path:
            def __init__(self, active_cell, path_cost, heluristic_cost, path):
                self.active_cell = active_cell
                self.path_cost = path_cost
                self.heluristic_cost = heluristic_cost
                self.path = path
            
            def get_total_cost(self):
                return self.path_cost + self.heluristic_cost

            def spawn_new_path(self, new_active_cell, new_heluristic_cost):
                path_copy = copy.copy(self.path)
                path_copy.append(new_active_cell)
                return possible_path(new_active_cell, self.path_cost + 1, new_heluristic_cost, path_copy)
        
        evaluated_cells = {}
        unevaluated_cells = {}
        unevaluated_cells[self.current_cell] = possible_path(self.current_cell, 0, self.calc_heluristic(self.current_cell, goals),[])
        
        while (len(unevaluated_cells) > 0):

            if len(unevaluated_cells) is 1:
                (cell, path_to_examine) = unevaluated_cells.popitem()
            else:
                cell = min(unevaluated_cells.keys(), key=lambda k1: unevaluated_cells[k1].get_total_cost())
                path_to_examine = unevaluated_cells.pop(cell)

            ajacent = path_to_examine.active_cell.get_adjacent_cell_map().values()
            for a_cell in ajacent:

                if path_to_examine.active_cell.has_fence_between(a_cell):
                    continue

                if not a_cell.get_passable():
                    continue
                
                new_path = path_to_examine.spawn_new_path(a_cell, self.calc_heluristic(a_cell, goals))
                
                if self.check_goals(a_cell, goals):
                    return new_path.path
                
                if a_cell in evaluated_cells.keys():
                    old_path = evaluated_cells[a_cell]
                    if old_path.path_cost > new_path.path_cost:
                        evaluated_cells[a_cell] = new_path
                    continue
                
                if a_cell in unevaluated_cells.keys():
                    old_path = unevaluated_cells[a_cell]
                    if old_path.path_cost > new_path.path_cost:
                        unevaluated_cells[a_cell] = new_path
                else:
                    unevaluated_cells[a_cell] = new_path
                    
            evaluated_cells[cell] = path_to_examine
            
        if(len(evaluated_cells) > 1):
            cell = min(evaluated_cells.keys(), key=lambda k1: evaluated_cells[k1].get_total_cost())
            best_path = evaluated_cells.pop(cell)
            return best_path.path
        elif(len(evaluated_cells) == 1):
            return evaluated_cells.values()[0].path
        else:
            return []
    
    def simulate(self, gamestate):
        #Degrade all goals (they will be enhanced if they are seen again this tick)
        for goal_obj in self.current_goals.keys():
            goal = self.current_goals[goal_obj]
            goal.tick()
            if goal.get_strength() <= 0:
                #forget about the goal
                self.current_goals.pop(goal_obj)
                self.goals_changed = True
        
        #Check my field of vision
        stuff_i_can_see = self.get_Field_Of_View()
        for seen_cell in stuff_i_can_see.keys():
            #update the world I know
            if seen_cell in self.known_world.keys():
                #Throw out what I knew about the old cell
                self.known_world.pop(seen_cell)
            self.known_world[seen_cell] = self.Mental_Cell(seen_cell,stuff_i_can_see[seen_cell])
            for obj in stuff_i_can_see[seen_cell]:
                #Update goals
                if obj in self.current_goals.keys():
                    goal = self.current_goals[obj]
                    old_loc = goal.Location
                    goal.Location = self.known_world[seen_cell]
                    if old_loc is not goal.Location:
                        self.goals_changed = True
                    goal.reset_age()
                else:
                    if obj.get_type() == PF_FLOWERBED:
                        if obj not in self.exhuasted_goals:
                            self.current_goals[obj] = self.Goal(AI_G_REACH, self.known_world[seen_cell], 1)
                            self.goals_changed = True
                    if obj.get_type() == PF_OLD_MAN:
                        if obj not in self.exhuasted_goals:
                            self.current_goals[obj] = self.Goal(AI_G_AVOID, self.known_world[seen_cell], 1)
                            self.goals_changed = True

        if (self.last_path is None) or (self.goals_changed):
            if (self.mental_state is AI_S_NORMAL) and (len(self.current_goals) > 0):
                path = self.pathfind(self.current_goals.values())
            else:
                path = self.pathfind([self.terminal_goal])
            self.last_path = path
            self.goals_changed = False

        if len(self.last_path) > 0:
            move_cell=self.last_path[0]
            if(self.move(move_cell)):
                self.last_path.remove(move_cell)
            else:
                #if for some reason the path fails,
                #forget about it
                self.last_path = None
        else:
            self.last_path = None
        
        for (goal_obj, goal) in self.current_goals.items():
            if goal.check_goals(self.current_cell):
                if goal.fufill():
                    self.current_goals.pop(goal_obj)
                    self.exhuasted_goals.append(goal_obj)
                    self.goals_changed = True

        # we are done, we can rest now
        if self.terminal_goal.check_goals(self.current_cell):
            self.current_cell.remove_object(self)

class person(AI_obj):
    def __init__(self, current_cell, terminal_cell):
        AI_obj.__init__(self, PF_PERSON, current_cell, terminal_cell)

class cell:
    """
    A single cell in the playing field
    """
    def __init__(self, xy_cor, playfield):
        self.x, self.y = xy_cor
        self.playfield = playfield
        self.objects = []
        
        self.opacity = 1

    def __repr__(self):
        return "cell(" + str(self.x) + "," + str(self.y) + ")"

    def add_object(self, target_obj):
        self.objects.append(target_obj)
        assert(self.playfield.add_object(target_obj))
        return True

    def remove_object(self, target_obj):
        if target_obj in self.objects:
            self.objects.remove(target_obj)
            assert(self.playfield.remove_object(target_obj))
            return True
        return False

    def get_opacity(self):
        if len(self.objects) == 0:
            return 1
        else:
            return max( obj.get_opacity() for obj in self.objects )

    def get_passable(self):
        for obj in self.objects:
            if not obj.get_passable():
                return False
        return True

    def get_objects(self):
        return self.objects

    def get_distance(self, target_cell):
        return abs(self.x - target_cell.x) + abs(self.y - target_cell.y)

    def get_adjacent_cell_in_direction(self, direction):
        if direction not in D_DIRECTIONS:
            return None
        
        if direction is D_UP:
            x,y = (self.x ,self.y+1 )
        elif direction is D_LEFT:
            x,y = (self.x-1 ,self.y   )
        elif direction is D_DOWN:
            x,y = (self.x   ,self.y-1 )
        elif direction is D_RIGHT:
            x,y = (self.x+1 ,self.y   )

        if ( 0 <= x < PLAYFIELD_X_SIZE) and ( 0 <= y < PLAYFIELD_Y_SIZE):
            return self.playfield.get_cell((x,y))
        else:
            return None


    def get_adjacent_cell_map(self):
        adjacent_cor = {
            D_UP:           (self.x   ,self.y+1 ),
            D_LEFT:         (self.x-1 ,self.y   ),
            D_DOWN:         (self.x   ,self.y-1 ),
            D_RIGHT:        (self.x+1 ,self.y   )
        }

        adjacent_cell = {}
        for direction in adjacent_cor.keys():
            x,y = adjacent_cor[direction]
            if ( 0 <= x < PLAYFIELD_X_SIZE) and ( 0 <= y < PLAYFIELD_Y_SIZE):
                adjacent_cell[direction] = self.playfield.get_cell((x,y))
        
        return adjacent_cell

    def has_fence_between(self, other_cell):
        if other_cell is None:
            return False
        return self.playfield.has_fence( ((self.x, self.y), (other_cell.x, other_cell.y)) )

    def get_LOS_matrix(self, Threshold = .05):
        """
        Generates the base LOS matrix used in the FOV function
        Takes the LOS_RAYS list (a list of all important sight rays)
        and maps them around this cell, generating occulsion values for
        each cell as it goes
        """
        matrix = {}
        for ray in LOS_RAYS:
            cumulative_occulsion = 1
            (prev_target_x, prev_target_y) = (self.x, self.y)
            for rx,ry in ray:
                target_x = rx + self.x
                target_y = ry + self.y
                if ( not(0 <= target_x < PLAYFIELD_X_SIZE) or not( 0 <= target_y < PLAYFIELD_Y_SIZE)):
                    break
                cell = self.playfield.get_cell((target_x,target_y))
                occlusion = (1-VIEW_DROPOFF*(abs(rx)+abs(ry)-1)) * cumulative_occulsion
                if occlusion < Threshold:
                    break
                if cell in matrix.keys():
                    if occlusion > matrix[cell]:
                        matrix[cell] = occlusion
                else:
                    matrix[cell] = occlusion

                cumulative_occulsion *= cell.get_opacity()
                if self.playfield.has_fence( ((prev_target_x, prev_target_y), (target_x, target_y)) ):
                    cumulative_occulsion *= cell.get_opacity()

        return matrix
        
class playfield:
    """
    A collection of cells arranged in a grid
    """
    def __init__(self):
        self.field = {}
        self.objects = []
        self.fences = sets.Set()

        #Create a the playfield and populate it with cells
        for x_cor in range(PLAYFIELD_X_SIZE):
            for y_cor in range(PLAYFIELD_Y_SIZE):
                self.field[(x_cor,y_cor)] = cell((x_cor,y_cor), self)

        #Iterate over the initial setup dict and put everying
        #in it's starting place
        for xy_cor in PLAYFIELD_INITIAL_SETUP.keys():
            obj_type = PLAYFIELD_INITIAL_SETUP[xy_cor]
            if obj_type == PF_EMPTY:
                continue

            target_cell = self.field[xy_cor]
            if obj_type == PF_HOUSE:
                obj = house_obj(target_cell)
            elif obj_type == PF_WINDOW:
                obj = window_obj(target_cell)
            elif obj_type == PF_DOOR:
                obj = door_obj(target_cell)
            elif obj_type == PF_SIDEWALK:
                obj = sidewalk_obj(target_cell)
            elif obj_type == PF_MAILBOX:
                obj = mailbox_obj(target_cell)
            elif obj_type == PF_SHED:
                obj = shed_obj(target_cell)
            elif obj_type == PF_OLD_MAN:
                obj = old_man_obj(target_cell)
            target_cell.add_object(obj)

        for fence_def in PLAYFIELD_INITIAL_FENCES:
            self.fences.add( sets.Set(fence_def) )

    def add_object(self, target_obj):
        self.objects.append(target_obj)
        target_obj.set_existance()
        return True

    def remove_object(self, target_obj):
        if target_obj in self.objects:
            self.objects.remove(target_obj)
            target_obj.clear_existance()
            return True
        return False

    def simulate(self, gamestate):
        for obj in self.objects:
            obj.simulate(gamestate)

    def get_cell(self, xy_cor):
        return self.field[xy_cor]

    def has_fence(self, fence_def_to_find):
        fence_to_find = sets.Set(fence_def_to_find)
        return fence_to_find in self.fences


### simple testhook
"""
if __name__ == '__main__':
    g = GameState()
    g.create_playfield()
    p = g.playfield
    c = p.get_cell((9,9))
    t = p.get_cell((0,0))
    obj = person(c,t)
    c.add_object(obj)

    while(obj.exists):
        g.simulate()
        print(obj.current_cell)
    print("Done")
"""